import { Request, Response } from "express";
import Cache from "node-cache";
import { Route, RouteMessages } from "server";
import { nodeManager } from "..";

const IGNORED_PATHS = ["/favicon.ico"];

const cache = new Cache({
  stdTTL: 300, // 5 minutes
});
type CachedRequest = {
  nodeId: string;
  status: number;
  headers: any;
  data: any;
};

/**
 * Logs a request
 *
 * @param nodeId the node ID that handled the request
 * @param url the URL of the request
 * @param status the status code of the request
 * @param time the time it took to handle the request (or true if it was cached)
 */
function log(
  nodeId: string,
  url: string,
  status: number,
  time: number | boolean
) {
  console.log(
    `[node: ${nodeId}] ${url} - ${status} (${
      time === true ? "cached" : time + "ms"
    })`
  );
}

export default class ProxyRoute extends Route {
  constructor() {
    super({ path: "/" });
  }

  async handle(req: Request, res: Response) {
    const url = req.url.substring(1);
    if (!url) {
      res.status(400).json(RouteMessages.badRequest("No URL provided"));
      return;
    }
    if (IGNORED_PATHS.includes(url)) {
      res.status(404).send("Not found");
      return;
    }

    const node = nodeManager.getRandomNode();
    if (!node) {
      res
        .status(500)
        .json(RouteMessages.internalServerError("No nodes available"));
      return;
    }

    try {
      const cachedRequest = cache.get<CachedRequest>(url);
      if (cachedRequest) {
        log(cachedRequest.nodeId, url, cachedRequest.status, true);
        res
          .status(cachedRequest.status)
          .set(cachedRequest.headers)
          .json(cachedRequest.data);
        return;
      }

      const before = Date.now();
      const response = await node.fetch(url);
      const nodeId = response.headers["x-proxy-node"];
      const data = response.data;

      if (response.status === 500) {
        res.status(500).json(RouteMessages.internalServerError(data));
        return;
      }

      log(nodeId, url, response.status, Date.now() - before);

      // Return the JSON response
      res.status(response.status).set(response.headers).json(data);
      cache.set(url, {
        nodeId: nodeId,
        status: response.status,
        headers: response.headers,
        data: data,
      } as CachedRequest);
    } catch (ex: any) {
      res.status(500).json(RouteMessages.internalServerError(ex.message || ex));
    }
  }
}
